# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import defaultdict
from typing import Any, Literal, Optional, Union

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype
from pydantic import BaseModel, Field, validator
from sapientml.params import Code, Config, Task

from .meta_features import (
    MetaFeatures,
    generate_column_meta_features,
    generate_model_meta_features,
    generate_pp_meta_features,
)

PipelineSkeleton = (
    dict  # dict[str, Union[float, dict[str, Union[float, list[str], list[dict[str, Union[float, int, str]]]]]]]
)

MAX_NUM_OF_COLUMNS = 10000000
MAX_SEED = 2**32 - 1
MAX_COLUMN_NAME_LENGTH = 1000
MAX_HPO_N_TRIALS = 100000
MAX_HPO_TIMEOUT = 500000
MAX_N_MODELS = 30
INITIAL_TIMEOUT = 600


class SapientMLConfig(Config):
    n_models: int = 3
    seed_for_model: int = 42
    id_columns_for_prediction: Optional[list[str]] = None
    use_word_list: Optional[Union[list[str], dict[str, list[str]]]] = None
    split_stratify: Optional[bool] = None
    use_hyperparameters: bool = False
    impute_all: bool = True
    hyperparameter_tuning: bool = False
    hyperparameter_tuning_n_trials: int = 10
    hyperparameter_tuning_timeout: int = 0
    hyperparameter_tuning_random_state: int = 1023
    predict_option: Literal["default", "probability"] = "default"
    permutation_importance: bool = True
    add_explanation: bool = False

    def postinit(self):
        if self.id_columns_for_prediction is None:
            self.id_columns_for_prediction = []

        if self.initial_timeout is None:
            if self.hyperparameter_tuning:
                self.initial_timeout = 0
            else:
                self.initial_timeout = INITIAL_TIMEOUT
            if self.hyperparameter_tuning_timeout is None:
                self.hyperparameter_tuning_timeout = INITIAL_TIMEOUT
        elif self.hyperparameter_tuning_timeout is None:
            if self.hyperparameter_tuning:
                self.hyperparameter_tuning_timeout = self.initial_timeout
            else:
                self.hyperparameter_tuning_timeout = INITIAL_TIMEOUT

        if self.use_pos_list is None:
            self.use_pos_list = []

    @validator("n_models")
    def check_n_models(cls, v):
        if v <= 0 or MAX_N_MODELS < v:
            raise ValueError(f"{v} is out of [1, {MAX_N_MODELS}]")
        return v

    @validator(
        "id_columns_for_prediction",
        "use_word_list",
    )
    def check_num_of_column_names(cls, v):
        if v is None:
            return v
        if len(v.keys() if isinstance(v, dict) else v) >= MAX_NUM_OF_COLUMNS:
            raise ValueError(f"The number of columns must be smaller than {MAX_NUM_OF_COLUMNS}")
        return v

    @validator(
        "id_columns_for_prediction",
        "use_word_list",
    )
    def check_column_name_length(cls, v):
        if v is None:
            return v
        for _v in v.keys() if isinstance(v, dict) else v:
            if len(_v) >= MAX_COLUMN_NAME_LENGTH:
                raise ValueError(f"Column name length must be shorter than {MAX_COLUMN_NAME_LENGTH}")
        return v

    @validator("seed_for_model")
    def check_seed(cls, v):
        if v < 0 or MAX_SEED < v:
            raise ValueError(f"{v} is out of [0, {MAX_SEED}]")
        return v

    @validator("hyperparameter_tuning_n_trials")
    def check_hyperparameter_tuning_n_trials(cls, v):
        if v < 1 or MAX_HPO_N_TRIALS < v:
            raise ValueError(f"{v} is out of [1, {MAX_HPO_N_TRIALS}]")
        return v

    @validator("hyperparameter_tuning_timeout")
    def check_hyperparameter_tuning_timeout(cls, v):
        if v < 0 or MAX_HPO_TIMEOUT < v:
            raise ValueError(f"{v} is out of [0, {MAX_HPO_TIMEOUT}]")
        return v

    @validator("hyperparameter_tuning_random_state")
    def check_hyperparameter_tuning_random_state(cls, v):
        if v < 0 or MAX_SEED < v:
            raise ValueError(f"{v} is out of [0, {MAX_SEED}]")
        return v


class Column(BaseModel):
    dtype: str
    meta_features: Optional[MetaFeatures]
    has_negative_value: bool

    @validator("dtype")
    def check_dtype(cls, v):
        if len(v) > 100:
            raise ValueError(f"'{v}' is invalid as a dtype")
        return v

    @validator("meta_features")
    def check_meta_features(cls, v):
        if v is None:
            return v
        for k, _v in v.items():
            if not re.match(r"feature:[a-z_0-9]+", k):
                raise ValueError(f"'{k}' is invalid as a feature name")
            if isinstance(_v, int):
                pass
            elif isinstance(_v, float):
                pass
            elif isinstance(_v, str) and len(_v) >= MAX_COLUMN_NAME_LENGTH:
                raise ValueError(f"Meta feature string value must be shorter than {MAX_COLUMN_NAME_LENGTH}")
            elif isinstance(_v, list):
                for s in _v:
                    if not isinstance(s, str):
                        raise ValueError("The list must contain string values only")
                    elif len(s) >= MAX_COLUMN_NAME_LENGTH:
                        raise ValueError(
                            f"Column name length in meta features must be shorter than {MAX_COLUMN_NAME_LENGTH}"
                        )
        return v


class DatasetSummary(BaseModel):
    columns: dict[str, Column]
    meta_features_pp: MetaFeatures
    meta_features_m: MetaFeatures
    has_multi_class_targets: bool
    has_inf_value_targets: bool
    cols_almost_missing_string: Optional[list[str]] = None
    cols_almost_missing_numeric: Optional[list[str]] = None

    @validator("columns", "cols_almost_missing_string", "cols_almost_missing_numeric")
    def check_num_of_columns(cls, v):
        if v is None:
            return v
        if len(v.keys() if isinstance(v, dict) else v) >= MAX_NUM_OF_COLUMNS:
            raise ValueError(f"The number of columns must be smaller than {MAX_NUM_OF_COLUMNS}")
        return v

    @validator("columns", "cols_almost_missing_string", "cols_almost_missing_numeric")
    def check_column_name_length(cls, v):
        if v is None:
            return v
        for _v in v.keys() if isinstance(v, dict) else v:
            if len(_v) >= MAX_COLUMN_NAME_LENGTH:
                raise ValueError(f"Column name length must be shorter than {MAX_COLUMN_NAME_LENGTH}")
        return v

    @validator("meta_features_pp", "meta_features_m")
    def check_meta_features(cls, v):
        for k, _v in v.items():
            if not re.match(r"feature:[a-z_0-9]+", k):
                raise ValueError(f"'{k}' is invalid as a feature name")
            if isinstance(_v, int):
                pass
            elif isinstance(_v, float):
                pass
            elif isinstance(_v, str) and len(_v) >= MAX_COLUMN_NAME_LENGTH:
                raise ValueError(f"Meta feature string value must be shorter than {MAX_COLUMN_NAME_LENGTH}")
            elif isinstance(_v, list):
                for s in _v:
                    if not isinstance(s, str):
                        raise ValueError("The list must contain string values only")
                    elif len(s) >= MAX_COLUMN_NAME_LENGTH:
                        raise ValueError(
                            f"Column name length in meta features must be shorter than {MAX_COLUMN_NAME_LENGTH}"
                        )
        return v


class ModelLabel(BaseModel):
    label_name: str
    predict_proba: bool = False
    hyperparameters: Optional[Any] = None
    meta_features: list[Any] = Field(default_factory=list)

    def __str__(self):
        return self.label_name

    def __repr__(self):
        return str(self)


class Pipeline(Code):
    task: Task
    dataset_summary: DatasetSummary
    config: SapientMLConfig
    adaptation_metric: Optional[str] = None
    all_columns_datatypes: dict = Field(default_factory=dict)
    model: Optional[ModelLabel] = None
    inverse_target: bool = False
    sparse_matrix: bool = False  # Whether the data is converted to sparse matrix in the pipeline
    train_column_names: list[str] = Field(default_factory=list)
    test_column_names: list[str] = Field(default_factory=list)
    train_ignore_columns: list[str] = Field(default_factory=list)
    test_ignore_columns: list[str] = Field(default_factory=list)

    # To handle following case;
    #   metrics : Accuracy
    #   task    : multi-class and multi-targets
    # because sklearn.accuracy_score doesn't support multi-class
    is_multi_class_multi_targets: bool = False

    id_columns_for_prediction: list[str] = Field(default_factory=list)
    output_dir_path: str = ""

    # pipeline json
    pipeline_json: dict = Field(default_factory=lambda: defaultdict(dict))

    labels: Optional[PipelineSkeleton] = None


class CoreResult(BaseModel):
    pipelines: list[Pipeline]
    labels: PipelineSkeleton


def summarize_dataset(df_train: pd.DataFrame, task: Task) -> DatasetSummary:
    """
    Args:
        - df_train      : Input dataset
        - task          : Task

    Returns:
        DatasetSummary
    """
    is_multi_classes: list[bool] = []
    for target in task.target_columns:
        is_multi_classes.append(len(df_train[target].unique()) > 1)
    has_multi_class_targets = all(is_multi_classes)

    has_inf_value_targets = bool(df_train[task.target_columns].isin([np.inf, -np.inf]).sum().sum() > 0)

    # handle almost_missing columns
    cols_all_missing = df_train.columns[df_train.isnull().all()]
    cols_almost_missing = df_train.columns[df_train.isna().sum() / len(df_train) > 0.8]
    cols_almost_missing = sorted(list(set(cols_almost_missing) - set(cols_all_missing)))
    cols_almost_missing_numeric = [col for col in cols_almost_missing if is_numeric_dtype(df_train[col])]
    cols_almost_missing_string = [col for col in cols_almost_missing if not is_numeric_dtype(df_train[col])]

    columns = dict()
    for column_name in df_train.columns:
        if column_name not in task.ignore_columns:
            meta_features = generate_column_meta_features(df_train[[column_name]])
        else:  # in case ignore_columns (columns.keys() be called by Adaptation class)
            meta_features = None

        columns[column_name] = Column(
            dtype=str(df_train[column_name].dtype),
            meta_features=meta_features,
            has_negative_value=bool(
                is_numeric_dtype(df_train[column_name]) and (df_train[[column_name]].values <= 0).any()
            ),
        )

    # Generate the meta-features
    meta_features_pp = generate_pp_meta_features(df_train, task.target_columns)

    is_clf_task = 1 if task.task_type == "classification" else 0
    meta_features_m = generate_model_meta_features(df_train, task.target_columns, is_clf_task)

    return DatasetSummary(
        columns=columns,
        meta_features_pp=meta_features_pp,
        meta_features_m=meta_features_m,
        has_multi_class_targets=has_multi_class_targets,
        has_inf_value_targets=has_inf_value_targets,
        cols_almost_missing_string=cols_almost_missing_string,
        cols_almost_missing_numeric=cols_almost_missing_numeric,
    )



# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import pandas as pd


class Pipeline_Explanation:
    Feature_exp_map = {
        "feature:missing_values_presence": "There are missing values among the columns with a score of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:max_normalized_stddev": "The maximum of standard deviation of the normalized data (among the columns) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_imbalance_score": "The imbalance score among target values of the columns, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_category_presence": "The category score of the columns is present with value of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_category_binary_presence": "There is a binary category among the columns with a score of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_category_small_presence": "There is a small number of categories (between 2 to 5) among the columns with a score of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_text_presence": "There is a text value that includes blank space among the columns with a score of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_date_presence": "The date presents among the columns with a score of **{actual_value}** which is {operator} {threshold}.",
        "feature:str_category_large_presence": "There is a large number of categories (between 2 to 5) among the columns with a score of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:max_normalized_mean": "The average of normalized values of the columns by considering only ['int16', 'int32', 'int64', 'float16', 'float32', 'float64'] type is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:normalized_variation_across_columns": "The average of normlized values of the columns by considering only ['int16', 'int32', 'int64', 'float16', 'float32', 'float64'] type is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:max_skewness": "The maximum of skewness (according to scipy.stats.skew) among the columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:not_basic_cols": "The number of type of columns are out of regular types (int, float, string, boolean) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:dominant_0.8": "The number of top relative normalized frequency of values (dividing all values by the sum of values) with greater than 0.8 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:dominant_0.9": "The number of top relative normalized frequency of values (dividing all values by the sum of values) with greater than 0.9 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_10-50_cols": "The ratio of missing values among the columns are between 1% to 10% with the number of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_50-90_cols": "The ratio of missing values among the columns are between 50% to 90% with the number of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_90-100_cols": "The ratio of missing values among the columns are between 90% to 100% with the number of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_0-10_cols": "The ratio of missing values among the columns are between 0% to 10% with the number of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_10-50_cols": "The ratio of missing values among the columns are between 1% to 10% with the number of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_10-50_cols": "The ratio of missing values among the columns are between 1% to 10% with the number of **{actual_value}** which is {operator} **{threshold}**.",
        "feature:negative_cols": "The number of negative values among numerical columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:median_0-1_cols": "The total number of median of the columns with less than 1 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:median_1-10_cols": "The total number of median of the columns between 1 to 10, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:median_10-100_cols": "The total number of median of the columns between 10 to 100, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:median_100-1000_cols": "The total number of median of the columns between 100 to 1,000, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:median_1000-10000_cols": "The total number of median of the columns between 1,000 to 10,000, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:median_10000_cols": "The total number of median of the columns over 10,000, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:iqr_0-1_cols": "The total number of values with interquartile range (IQR)<=1 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:iqr_1-10_cols": "The total number of values with 1<interquartile range (IQR)<=10 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:iqr_10-100_cols": "The total number of values with 10<interquartile range (IQR)<=100 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:iqr_100-1000_cols": "The total number of values with 100<interquartile range (IQR)<=1,000 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:iqr_1000-10000_cols": "The total number of values with 1000<interquartile range (IQR)<=10,000 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:iqr_10000_cols": "The total number of values with interquartile range (IQR)>10,000 is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_date": "The total number of columns with date type (at least 80% of values has a date type) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_catg": "The number of category column is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_num": "The number of numerical column (where at least 90% of values are numerical) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_other": "The number of nonnumerical and noncategorical column is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_str_catg": "The number of target string category colums, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_str_text": "The number of target text columns, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_str_other": "The number of nonnumerical and noncategorical columns, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_catg_num_max": "The maximum number of number of categorical columns, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_catg_num_min": "The minimum number of number of categorical columns, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:ttest_max": "The maximum number of identical columns (where T-Test of feature column > 0.01 against target columns), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:ttest_min": "The minimum number of identical columns (where T-Test of feature column > 0.01 against target columns), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:kstest_max": "The maximum number of identical distributions columns (where Kolmogorov-Smirnov test of feature column > 0.01 against target columns), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:kstest_min": "The minimum number of identical distributions columns (where Kolmogorov-Smirnov test of feature column > 0.01 against target columns), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:pearsonr_corr_max": "The maximum number of correlated columns against the target columns (where the Pearson correlation coefficient > 0.6), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:pearsonr_corr_min": "The minimum number of correlated columns against the target columns (where the Pearson correlation coefficient > 0.6), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:pearsonr_p_max": "The maximum number of correlated columns against the target columns (where the Pearson correlation p-value > 0.01), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:pearsonr_p_min": "The minimum number of correlated columns against the target columns (where the Pearson correlation o-value > 0.01), is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_kurtosis_max": "The maximum value of target kurtosis is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:target_kurtosis_min": "The minimum value of target kurtosis is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_values_special": 'The total number of special values that include ["?", "??", "-", "--", "---"] is **{actual_value}** which is {operator} **{threshold}**.',
        "feature:missing_values_rows": "The total number of missing values, is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:missing_values_named_columns": "The total number of columns with a title is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:avg_num_words": "The average number of words is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_unique_words": "The average number of unique words is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_catg_small": "The total number of small categories (the number of unique values < 20) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_catg_large": "The total number of large categories (more than 80% unique values) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_catg_binary": "The total number of binary categories (the number of unique values = 2) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_catg_small": "The total number of small numerical categories (the number of unique values < 20) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_catg_large": "The total number of large categories (more than 80% unique values) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_text": "The ratio of text columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_catg": "The number of category columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_cont": "The number of columns with continues values is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:str_catg_binary": "The number of binary categories is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:outlier_cols": "The number of outlier columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:many_outlier_cols": "The ratio of outlier columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:kurtosis_normal": "The total number of columns with normal Kurtosis (kurtosis<0.5) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:kurtosis_uniform": "The total number of columns with normal Kurtosis (kurtosis<-1.0) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:kurtosis_tailed": "The total number of columns with uniform Kurtosis (kurtosis>=0.5) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:dist_normal": "The total number of columns with normal distribution (Kolmogorov-Smirnov test of feature with normal distribution < 0.5) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:dist_uniform": "The total number of columns with uniform distribution (Kolmogorov-Smirnov test of feature with uniform distribution < -1.0) is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:dist_poisson": "The total number of columns with poisson distribution (Kolmogorov-Smirnov test of feature with poisson distribution >= 0.5) is **{actual_value}** which is {operator} **{threshold}**.",
        "correlated_cols": "The total number of correlated columns is **{actual_value}** which is {operator} **{threshold}**.",
        "feature:num_of_rows": "The total number of observations (rows) is **{actual_value}** which is {operator} **{threshold}**.",
    }

    Operators = {
        ">": "greater than",
        "<": "smaller than",
        "=": "equal to",
        "<=": "smaller or equal to",
        ">=": "greater or equal to",
    }

    def __init__(self, skeleton: dict, explanation: dict, run_info: Optional[dict] = None):
        self.skeleton = skeleton
        self.explanation = explanation
        self.run_time = run_info

    def get_feature_explanation(self, action, cols, probability):
        explain = []
        counter = 0
        explain.append(
            "## ML Explainability Tips:\n\nThe following codes are generated with a confidence level of **{probability}** because of the following reasons apply to **{len_cols}** column(s) that include {cols}.\n\n".format(
                cols=cols, len_cols=len(cols), probability=probability
            )
        )
        for predicate in self.skeleton[action]["predicates"]:
            feature = predicate["feature_name"]
            operator = self.Operators[predicate["operator"]]
            threshold = predicate["threshold"]
            actual_value = predicate["actual_value"]
            counter += 1
            explain.append(
                "1. "
                + self.Feature_exp_map[feature].format(
                    cols=cols, operator=operator, threshold=round(threshold, 3), actual_value=round(actual_value, 3)
                )
            )
        return "\n".join(_ for _ in explain)

    @staticmethod
    def get_model_exp(model):
        model_expl = ""
        if model:
            if model.startswith("MODEL"):
                model_type = model.split(":")[1]
                model_name = model.split(":")[2]
                model_expl = f"## ML Explainability Tips:\n\nThis is a **{model_type}** problem and we are selecting **{model_name}** "
        return model_expl

    def process(self):
        explains = []
        ds = []
        for section in self.explanation:
            for action in self.explanation[section]:
                if "explanation" in self.explanation[section][action]:
                    code = self.explanation[section][action]["code"]
                    explain = self.explanation[section][action]["explanation"]
                    # meta = explain["relevant_meta_feature_list"]
                    probability = round(self.skeleton[action]["probability"], 2)
                    cols = explain["relevant_column_list"]
                    explains.append(
                        {
                            "code": code,
                            "action": action,
                            "explanation": self.get_feature_explanation(action, cols, probability),
                            "added": False,
                        }
                    )
            if section == "model":
                if "target_component_name" in self.explanation[section][action]:
                    model = self.explanation[section][action]["target_component_name"]
                    cols = self.explanation[section][action]["relevant_meta_feature_list"]
                    model_code = self.explanation[section]["code"]
                    prob = round(self.skeleton[model] / 2, 2)
                    explain_model = (
                        self.get_model_exp(model)
                        + f"with a probability of **{prob}**. The following are alternative options.\n"
                    )

                    metric = None  # in case run_time.json is not available
                    best_model_score = "N/A"
                    for key in self.skeleton.keys():
                        val = self.skeleton[key]
                        current_model = key.split(":")[2]
                        if key.startswith("MODEL:"):
                            score = "N/A"

                            if self.run_time:
                                for k in self.run_time:
                                    code_exec = self.run_time[k]
                                    if current_model.lower() in code_exec["content"]["model"]["label_name"].lower():
                                        if code_exec["run_info"]["score"]:
                                            score = round(code_exec["run_info"]["score"], 3)
                                            metric = code_exec["run_info"]["metric"]

                                        if model == key:  # top performed model
                                            best_model_score = score
                            if metric:
                                ds.append([current_model, f"{round(val/2,2)}", score])
                            else:
                                ds.append([current_model, f"{round(val/2,2)}"])
                    if metric:
                        ds_df = pd.DataFrame(ds, columns=["Model", "Probability", metric])
                    else:
                        ds_df = pd.DataFrame(ds, columns=["Model", "Probability"])
                    if ds_df.shape[0]:
                        ds_df.index = ds_df.index + 1
                        cols = []
                        explains.append(
                            {
                                "code": [model_code],
                                "action": model,
                                "explanation": explain_model  # basic_explain+
                                + (
                                    f"This model is selected based on the best **{metric}** of **{best_model_score}**.\n"
                                    if metric
                                    else ""
                                )
                                + ds_df.to_html(),
                                "added": False,
                            }
                        )
        return explains



# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import copy
import glob
import json
import re
from importlib.metadata import entry_points
from pathlib import Path
from shutil import copyfile
from typing import Tuple, Union

from sapientml.executor import PipelineExecutor
from sapientml.generator import CodeBlockGenerator, PipelineGenerator
from sapientml.macros import metric_lower_is_better
from sapientml.params import Code, Dataset, PipelineResult, RunningResult, Task
from sapientml.util.json_util import JSONEncoder
from sapientml.util.logging import setup_logger

from . import ps_macros
from .adaptation.generation.template_based_adaptation import Adaptation
from .explain.main import process as explain
from .params import Pipeline, SapientMLConfig, summarize_dataset
from .seeding.predictor import predict

model_dir_path_default = Path(__file__).parent / "models"

logger = setup_logger()


class SapientMLGenerator(PipelineGenerator, CodeBlockGenerator):
    def __init__(self, **kwargs):
        self.config = SapientMLConfig(**kwargs)
        self.config.postinit()
        eps = entry_points(group="sapientml.code_block_generator")
        self.loaddata = eps["loaddata"].load()(**kwargs)
        self.preprocess = eps["preprocess"].load()(**kwargs)

    def generate_pipeline(self, dataset: Dataset, task: Task):
        self.dataset = dataset
        self.task = task

        logger.info("Generating pipelines...")
        dataset, loaddata_block = self.loaddata.generate_code(dataset, task)
        dataset, preprocess_block = self.preprocess.generate_code(dataset, task)
        code_block = loaddata_block + preprocess_block
        dataset, sapientml_results = self.generate_code(dataset, task)

        result_pipelines: list[Code] = []
        for pipeline in sapientml_results:
            pipeline.validation = code_block.validation + pipeline.validation
            pipeline.test = code_block.test + pipeline.test
            pipeline.train = code_block.train + pipeline.train
            pipeline.predict = code_block.predict + pipeline.predict
            result_pipelines.append(pipeline)

        logger.info("Executing generated pipelines...")
        executor = PipelineExecutor()
        self.execution_results = executor.execute(
            result_pipelines,
            self.config.initial_timeout,
            Path(dataset.output_dir),
            self.config.cancel,
        )

        logger.info("Evaluating execution results of generated pipelines...")
        lower_is_better = self.task.adaptation_metric in metric_lower_is_better
        self.evaluate(self.execution_results, lower_is_better)
        logger.info("Done.")

        return (self._best_pipeline, self._best_pipeline_score), self._candidate_scripts

    def generate_code(self, dataset: Dataset, task: Task) -> Tuple[Dataset, list[Pipeline]]:
        df = dataset.training_dataframe
        # Generate the meta-features
        logger.info("Generating meta features ...")
        dataset_summary = summarize_dataset(df, task)  # type: ignore
        if dataset_summary.has_inf_value_targets:
            raise ValueError("Stopped generation because target columns have infinity value.")

        # discard columns with analysis
        # NOTE: The following code modify task.ignore_columns because ignore_columns is the same instance as task.ignore_columns.
        # 1. columns marked as STR_OTHER
        if ps_macros.STR_OTHER in dataset_summary.meta_features_pp:
            undetermined_column_names = dataset_summary.meta_features_pp[ps_macros.STR_OTHER]
            if isinstance(undetermined_column_names, list):
                task.ignore_columns += undetermined_column_names
        del dataset_summary.meta_features_pp[ps_macros.STR_OTHER]
        # 2. columns with all null values
        if ps_macros.ALL_MISSING_PRESENCE in dataset_summary.meta_features_pp:
            column_names_with_all_missing_values = dataset_summary.meta_features_pp[ps_macros.ALL_MISSING_PRESENCE]
            if isinstance(column_names_with_all_missing_values, list):
                task.ignore_columns += column_names_with_all_missing_values
        del dataset_summary.meta_features_pp[ps_macros.ALL_MISSING_PRESENCE]

        labels = predict(task, dataset_summary)
        adapt = Adaptation(
            labels,
            task,
            dataset_summary,
            self.config,
        )
        pipelines = adapt.run_adaptation()

        return dataset, pipelines

    def evaluate(self, execution_results: list[tuple[Code, RunningResult]], lower_is_better: bool = False) -> None:
        self._best_pipeline = None
        self._best_pipeline_score = PipelineResult(score=None, metric=None, best_params=None)
        candidate_scripts = []
        for pipeline, result in execution_results:
            if result.returncode == 0:
                pipeline_score = self._parse_pipeline_output(result.output)
            else:
                pipeline_score = PipelineResult(score=None, metric=None, best_params=None)
            candidate_scripts.append((pipeline, pipeline_score))
        self._candidate_scripts = candidate_scripts

        # When an error occurs while running a pipeline, the score becomes None
        error_pipelines = [pipeline for pipeline in candidate_scripts if pipeline[1].score is None]

        # If none of them have the score, stop ranking them
        if len(candidate_scripts) == len(error_pipelines):
            return

        # sort descending
        succeeded_scripts = sorted(
            [x for x in candidate_scripts if x[1].score is not None],
            key=lambda x: x[1].score,
            reverse=(not lower_is_better),
        )
        failed_scripts = [x for x in candidate_scripts if x[1].score is None]

        ranked_candidate_scripts = succeeded_scripts + failed_scripts
        best_pipeline_tuple = ranked_candidate_scripts[0]
        if best_pipeline_tuple is None:
            return

        best_pipeline = copy.deepcopy(best_pipeline_tuple[0])
        if best_pipeline_tuple[1].best_params is not None:
            best_pipeline.test = best_pipeline.test.replace(
                "best_params = study.best_params", "best_params = " + str(best_pipeline_tuple[1].best_params)
            )
            best_pipeline.train = best_pipeline.train.replace(
                "best_params = study.best_params", "best_params = " + str(best_pipeline_tuple[1].best_params)
            )
        self._best_pipeline = best_pipeline
        self._best_pipeline_score = best_pipeline_tuple[1]

    @staticmethod
    def _parse_pipeline_output(output: str):
        score = None
        best_params = None
        metric = None
        output_lines = output.splitlines()
        try:
            for line in output_lines:
                if re.match("best params: ", line):
                    best_params = ast.literal_eval(re.findall("best params: (.+)", line)[0])
                elif re.match("RESULT: ", line):
                    parts = [x.strip() for x in line.split(":")]
                    metric = parts[-2].strip().split(" ")[0]
                    score = float(parts[-1])
        except Exception:
            pass
        return PipelineResult(score=score, metric=metric, best_params=best_params)

    def save(self, output_dir: Union[Path, str]):
        if self.config.dry_run:
            return

        def add_prefix(filename, prefix):
            if not prefix:
                return filename
            return f"{prefix}_{filename}"

        path = Path(output_dir)
        path.mkdir(parents=True, exist_ok=True)

        _output_dir = Path(self.dataset.output_dir)

        candidate_scripts = self._candidate_scripts
        if candidate_scripts:
            if self._best_pipeline:
                script_body = self._best_pipeline.test.replace(_output_dir.as_posix(), ".")
                with open(path / add_prefix("final_script.py", self.config.project_name), "w", encoding="utf-8") as f:
                    f.write(script_body)

                script_body = self._best_pipeline.train.replace(_output_dir.as_posix(), ".")
                with open(path / add_prefix("final_train.py", self.config.project_name), "w", encoding="utf-8") as f:
                    f.write(script_body)

                script_body = self._best_pipeline.predict.replace(_output_dir.as_posix(), ".")
                with open(path / add_prefix("final_predict.py", self.config.project_name), "w", encoding="utf-8") as f:
                    f.write(script_body)

                with open(
                    path / (add_prefix("final_script", self.config.project_name) + ".out.json"),
                    "w",
                    encoding="utf-8",
                ) as f:
                    json.dump(self._best_pipeline_score.__dict__, f, cls=JSONEncoder, indent=4)
            else:
                logger.warning("All candidate scripts failed. Final script is not saved.")
                raise RuntimeError("All candidate scripts failed. Final script is not saved.")

            # copy libs
            lib_path = path / "lib"
            lib_path.mkdir(exist_ok=True)

            eps = entry_points(group="sapientml.export_modules")
            for ep in eps:
                for file in glob.glob(f"{ep.load().__path__[0]}/*.py"):
                    copyfile(file, lib_path / Path(file).name)

            for index, (script, detail) in enumerate(candidate_scripts, start=1):
                # script.dataset.training_data_path is '{user specified dir}/{name}.csv' or '{tmpdir}/training.pkl'
                # If latter one, we have to modify the {tmpdir} to output_dir.
                script_body = script.validation.replace(_output_dir.as_posix(), ".")

                with open(path / f"{index}_script.py", "w", encoding="utf-8") as f:
                    f.write(script_body)

        skeleton = self._best_pipeline.labels

        debug_info = {}
        for i, candidate in enumerate(candidate_scripts, start=1):
            info = {"content": candidate[0].dict(), "run_info": candidate[1].__dict__}
            debug_info[i] = info

        if self.config.debug:
            with open(path / add_prefix("run_info.json", self.config.project_name), "w", encoding="utf-8") as f:
                json.dump(debug_info, f, cls=JSONEncoder, indent=4)

        if self.config.add_explanation:
            explain(
                visualization=True,
                eda=True,
                dataframe=self.dataset.training_dataframe,
                script_path=(Path(output_dir) / add_prefix("final_script.py", self.config.project_name))
                .absolute()
                .as_posix(),
                target_columns=self.task.target_columns,
                problem_type=self.task.task_type,
                ignore_columns=self.task.ignore_columns,
                skeleton=skeleton,
                explanation=self._best_pipeline.pipeline_json,
                run_info=debug_info,
                internal_execution=True,
                timeout=self.config.timeout_for_test,
                cancel=self.config.cancel,
            )
